#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "cycles.h"
#include "rnlp.h"

/*
 *  RNLP: initialize the system
 */
void rnlp_init(rnlp_lock_t* lock)
{
    lock->head = NULL;
    lock->tail = NULL;
    lock->locked = 0;
    lock->unavailable = 0;
    ticket_init(&(lock->sublock));
}

/*
 *  RNLP: lock
 */
uint64_t *rnlp_lock(rnlp_lock_t* lock, rnlp_node_t* self)
{
    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    ticket_lock(&(lock->sublock));

    if (0 == (lock->unavailable & self->requested))
    {
        self->blocked = 0;
        lock->locked |= self->requested;
        lock->unavailable |= self->requested;
    }
    else if (NULL == lock->head)
    {
        self->blocked = 1;
        lock->head = self;
        lock->tail = self;
        self->next = NULL;
    }
    else
    {
        self->blocked = 1;
        lock->tail->next = self;
        self->next = NULL;
        lock->tail = self;
    }

    ticket_unlock(&(lock->sublock));

    end = get_cycles();
    ret[0] += end - start;
    start = get_cycles();

    while (self->blocked == 1)
    {
        cpu_relax();
    }

    end = get_cycles();
    ret[1] += end - start;

    return ret;
}

/*
 *  RNLP: unlock
 */
uint64_t *rnlp_unlock(rnlp_lock_t* lock, rnlp_node_t* self)
{
    rnlp_node_t *tmp, *prev = NULL;
    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    ticket_lock(&(lock->sublock));

    lock->locked &= ~(self->requested);
    lock->unavailable = lock->locked;

    tmp = (rnlp_node_t*)lock->head;

    while (tmp != NULL)
    {
        assert(tmp->blocked != 0);

        if (0 == (lock->unavailable & tmp->requested))
        {
            tmp->blocked = 0;
            lock->locked |= tmp->requested;

            lock->unavailable |= tmp->requested;

            if (tmp == lock->head)
            {
                lock->head = tmp->next;
                tmp = tmp->next;
            }
            else if (tmp == lock->tail)
            {
                prev->next = NULL;
                lock->tail = prev;
                tmp = NULL;
            }
            else
            {
                prev->next = tmp->next;
                tmp = tmp->next;
            }
        }
        else
        {
            lock->unavailable |= tmp->requested;
            prev = tmp;
            tmp = tmp->next;
        }
    }

    ticket_unlock(&(lock->sublock));

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

